"""Async Spa Structure block."""

from __future__ import annotations

import asyncio
import datetime
import importlib
import logging
from functools import partial
from typing import TYPE_CHECKING, Any

from geckolib._version import VERSION
from geckolib.config import GeckoConfig
from geckolib.const import GeckoConstants
from geckolib.driver.accessor import (
    GeckoByteStructAccessor,
)

INMIX_PACKTYPE = 14

if TYPE_CHECKING:
    from collections.abc import Callable
    from types import ModuleType

    from geckolib.driver.async_udp_protocol import GeckoAsyncUdpProtocol

_LOGGER = logging.getLogger(__name__)


class GeckoAsyncStructure:
    """Class to host/manage the raw data block for a spa structure."""

    def __init__(self, on_async_set_value: Callable) -> None:
        """Initialize the async version."""
        self._on_async_set_value = on_async_set_value
        self.reset()

    def replace_status_block_segment(self, offset: int, segment: bytes) -> None:
        """Replace a segment of the status block."""
        previous_block = self._status_block
        segment_len = len(segment)
        self._status_block = (
            self._status_block[0:offset]
            + segment
            + self._status_block[offset + segment_len :]
        )
        # Notify changes to accessors
        for accessor in self.accessors.values():
            accessor.status_block_changed(offset, segment_len, previous_block)

    @property
    def status_block(self) -> bytes:
        """Get the status block."""
        return self._status_block

    def set_status_block(self, block: bytes) -> None:
        """Set the status block."""
        self._status_block = block

    def build_accessors(self) -> None:
        """Build the accessors."""
        self.accessors = dict(
            self.config_class.accessors,
            **self.log_class.accessors,
            **self.inmix_config_accessors,
            **self.inmix_log_accessors,
        )
        # Get all outputs
        self.all_outputs = [*self.config_class.output_keys, *self.log_class.output_keys]
        # Get collection of possible devices
        self.all_devices = self.log_class.all_device_keys
        # User devices are those that have a Ud in the tag name
        self.user_demands = self.log_class.user_demand_keys
        # Error keys
        self.error_keys = self.log_class.error_keys

    def reset(self) -> None:
        """Reset this status block to initialization state."""
        self.accessors = {}
        self.all_outputs = []
        self.all_devices = []
        self.user_demands = []
        self.connections = []
        self._status_block: bytes = b"\x00" * 1024

        self.plateform_key: str = ""
        self.pack_class = None
        self.pack_type: int = 0
        self.config_version: int = 0
        self.config_class = None
        self.log_version: int = 0
        self.log_class = None

        self.inmix_config_class = None
        self.inmix_config_accessors = {}
        self.inmix_log_class = None
        self.inmix_log_accessors = {}

    async def get(
        self,
        protocol: GeckoAsyncUdpProtocol,
        create_func: Callable,
        packet_timeout: int = GeckoConfig.PAUSE_BETWEEN_RETRIES_IN_SECONDS,
    ) -> bool:
        """Get response from a command."""
        _LOGGER.debug("Async get for struct")
        async with protocol.lock:
            # Create the request
            request = create_func()

            try:
                async with asyncio.timeout(request.timeout_in_seconds):
                    while True:
                        # Queue it for delivery
                        protocol.queue_send(request)

                        # Setup some controls for the multiple responses expected
                        next_expected = 0
                        segments = []

                        while True:
                            # Wait for a response up to a certain amount of time
                            if await request.wait_for_response(
                                protocol, packet_timeout
                            ):
                                _LOGGER.debug("timeout waiting for any block")
                                break

                            if next_expected == request.sequence:
                                segments.append(request.data)
                                next_expected = request.next

                                if request.next == 0:
                                    _LOGGER.debug(
                                        "Status block segments complete, update"
                                    )
                                    self.replace_status_block_segment(
                                        request.start,
                                        b"".join(segments),
                                    )
                                    return True
                            else:
                                _LOGGER.debug(
                                    "Out-of-sequence status block segment %d - ignored",
                                    request.sequence,
                                )

                                if request.next == 0:
                                    break

                        # Create a clean new request
                        request = create_func()

            except TimeoutError:
                _LOGGER.debug(
                    "TIMEOUT: Handler %s threw TimeoutError (%fs)",
                    request,
                    request.timeout_in_seconds,
                )

            return False

    async def async_set_value(self, pos: int, length: int, newvalue: Any) -> None:
        """Set the value of a block."""
        await self._on_async_set_value(pos, length, newvalue)

    async def _async_import_module(self, module_name: str) -> ModuleType:
        """Load a module asyncronously."""
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(
            None, partial(importlib.import_module, module_name)
        )

    def get_differences(self, other: GeckoAsyncStructure) -> list[str]:
        """Get the differences between this structure and the other."""
        results = []
        if len(self.accessors) != len(other.accessors):
            results.append(
                f"Accessor length difference. I have {len(self.accessors)},"
                f" the other has {len(other.accessors)}"
            )

        # List my unique accessors
        unique_to_me = [
            self.accessors[key] for key in self.accessors if key not in other.accessors
        ]
        results.extend([f"Only me: {acc}" for acc in unique_to_me])

        # List other unique accessors
        unique_to_other = [
            other.accessors[key] for key in other.accessors if key not in self.accessors
        ]
        results.extend([f"Only other: {acc}" for acc in unique_to_other])

        # List shared accessor differences
        key_differences = [
            key
            for key in self.accessors
            if key in other.accessors and self.accessors[key] != other.accessors[key]
        ]
        results.extend(
            [
                f"Diff: {self.accessors[key]} != {other.accessors[key]}"
                for key in key_differences
            ]
        )

        return results

    ############################################################################
    #
    #   DRY functionality

    async def load_pack_class(self, plateform_key: str) -> None:
        """Load the pack class for this structure."""
        self.plateform_key = plateform_key
        pack_module_name = f"geckolib.driver.packs.{plateform_key}"
        pack_class = (await self._async_import_module(pack_module_name)).GeckoPack
        self.pack_class = pack_class(self)
        self.pack_type = self.pack_class.plateform_type

    async def load_config_module(self, config_version: int) -> None:
        """Load the config pack class for this structure."""
        config_module_name = (
            f"geckolib.driver.packs.{self.plateform_key}-cfg-{config_version}"
        )
        config_class = (
            await self._async_import_module(config_module_name)
        ).GeckoConfigStruct
        self.config_version = config_version
        self.config_class = config_class(self)

    async def load_log_module(self, log_version: int) -> None:
        """Load the log pack class for this structure."""
        log_module_name = (
            f"geckolib.driver.packs.{self.plateform_key}-log-{log_version}"
        )
        log_class = (await self._async_import_module(log_module_name)).GeckoLogStruct
        self.log_version = log_version
        self.log_class = log_class(self)

    def get_snapshot_data(self) -> dict:
        """Get snapshot data that the structure can supply."""
        pack_id = GeckoConstants.KEY_PACK_CONFIG_ID
        if pack_id not in self.accessors:
            pack_id = GeckoConstants.KEY_PACK_CORE_ID
        rev_id = GeckoConstants.KEY_PACK_CONFIG_REV
        if rev_id not in self.accessors:
            rev_id = GeckoConstants.KEY_PACK_CORE_REV
        rel_id = GeckoConstants.KEY_PACK_CONFIG_REL
        if rel_id not in self.accessors:
            rel_id = GeckoConstants.KEY_PACK_CORE_REL

        return {
            "Library Version": VERSION,
            "SpaPackStruct.xml revision": self.pack_class.revision
            if self.pack_class is not None
            else "Not Known",
            "Spa pack": f"{
                self.accessors[GeckoConstants.KEY_PACK_TYPE].value
                if GeckoConstants.KEY_PACK_TYPE in self.accessors
                else ' (NO PACK TYPE)'
            }"
            f" {
                self.accessors[pack_id].value
                if pack_id in self.accessors
                else ' (NO PACK)'
            }"
            f" v{
                self.accessors[rev_id].value
                if rev_id in self.accessors
                else ' (NO REV)'
            }"
            f".{
                self.accessors[rel_id].value
                if rel_id in self.accessors
                else ' (NO REL)'
            }",
            "Low level configuration #": self.accessors[
                GeckoConstants.KEY_CONFIG_NUMBER
            ].value
            if GeckoConstants.KEY_CONFIG_NUMBER in self.accessors
            else "(NO CONFIG #)",
            "Log version": self.log_version,
            "Config version": self.config_version,
            "Pack type": self.pack_type,
            "Snapshot UTC Time": f"{datetime.datetime.now(tz=datetime.UTC)}",
            "Status Block": [hex(b) for b in self.status_block],
        }

    def build_connections(self) -> None:
        """Scan the spa pack outputs to generate a list of connections."""
        _LOGGER.debug("All outputs are %s", self.all_outputs)

        self.connections = [
            value
            for output in self.all_outputs
            if (value := self.accessors[output].value) not in ["NA", "Not_Set"]
        ]

        _LOGGER.debug("Connections are %s", self.connections)

    ############################################################################
    #
    #   Accessory support

    @property
    def is_mr_steam(self) -> bool:
        """Is this a MrSteam pack."""
        return self.plateform_key == "mrsteam"

    @property
    def is_bain_ultra(self) -> bool:
        """Is this a Bain Ultra pack."""
        return self.plateform_key == "mas-ibc-32k"

    @property
    def is_spa_pack(self) -> bool:
        """Is this a regular pack pack."""
        return not (self.is_mr_steam or self.is_bain_ultra)

    async def check_for_accessories(self) -> None:
        """After the initial pack has been loaded, check for accessories too."""
        inmix_packtype = GeckoByteStructAccessor(self, "inMix-PackType", 628, None)
        if inmix_packtype.value == INMIX_PACKTYPE:
            inmix_configlib = GeckoByteStructAccessor(
                self, "inMix-ConfigLib", 634, None
            )
            inmix_statuslib = GeckoByteStructAccessor(
                self, "inMix-StatusLib", 635, None
            )
            await self.load_inmix_config_module(inmix_configlib.value)
            await self.load_inmix_log_module(inmix_statuslib.value)

    async def load_inmix_config_module(self, config_version: int) -> None:
        """Load the config module for the inmix system."""
        config_module_name = f"geckolib.driver.packs.inmix-cfg-{config_version}"
        config_class = (
            await self._async_import_module(config_module_name)
        ).GeckoConfigStruct
        self.inmix_config_class = config_class(self)
        self.inmix_config_accessors = self.inmix_config_class.accessors

    async def load_inmix_log_module(self, log_version: int) -> None:
        """Load the config module for the inmix system."""
        log_module_name = f"geckolib.driver.packs.inmix-log-{log_version}"
        log_class = (await self._async_import_module(log_module_name)).GeckoLogStruct
        self.inmix_log_class = log_class(self)
        self.inmix_log_accessors = self.inmix_log_class.accessors



"""GeckoConfigStruct - A class to manage the ConfigStruct for 'InXM v4'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoTempStructAccessor,
    GeckoTimeStructAccessor,
)


class GeckoConfigStruct:
    """Config Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the config struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the config struct class version."""
        return 4

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return [
            "Out1A",
            "Out1B",
            "Out2A",
            "Out2B",
            "Out3A",
            "OutHeater",
            "Out4A",
            "Out5A",
            "Out6A",
            "Out7A",
            "OutLi",
            "OutIO2",
            "OutIO4",
        ]

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "ConfigNumber": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/ConfigNumber", 0, "ALL"
            ),
            "Out1A": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out1A",
                95,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                ],
                None,
                None,
                "ALL",
            ),
            "Out1B": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out1B",
                96,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                ],
                None,
                None,
                "ALL",
            ),
            "Out2A": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out2A",
                97,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                ],
                None,
                None,
                "ALL",
            ),
            "Out2B": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out2B",
                98,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                ],
                None,
                None,
                "ALL",
            ),
            "Out3A": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out3A",
                99,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                ],
                None,
                None,
                "ALL",
            ),
            "OutHeater": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/OutHeater",
                104,
                None,
                ["NA", "", "", "", "", "", "", "", "", "", "", "", "", "", "HTR"],
                None,
                None,
                "ALL",
            ),
            "Out1ACurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out1ACurrent", 108, "ALL"
            ),
            "Out1BCurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out1BCurrent", 109, "ALL"
            ),
            "Out2ACurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out2ACurrent", 110, "ALL"
            ),
            "Out2BCurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out2BCurrent", 111, "ALL"
            ),
            "Out3ACurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out3ACurrent", 112, "ALL"
            ),
            "Out4A": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out4A",
                100,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                ],
                None,
                None,
                "ALL",
            ),
            "Out5A": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out5A",
                101,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                ],
                None,
                None,
                "ALL",
            ),
            "Out6A": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out6A",
                102,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                ],
                None,
                None,
                "ALL",
            ),
            "Out7A": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out7A",
                103,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                ],
                None,
                None,
                "ALL",
            ),
            "Out4ACurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out4ACurrent", 113, "ALL"
            ),
            "Out5ACurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out5ACurrent", 114, "ALL"
            ),
            "Out6ACurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out6ACurrent", 115, "ALL"
            ),
            "Out7ACurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out7ACurrent", 116, "ALL"
            ),
            "DirectCurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/DirectCurrent", 117, "ALL"
            ),
            "OutLi": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LVOutputConfig/OutLi",
                105,
                None,
                ["NA", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "LI"],
                None,
                None,
                "ALL",
            ),
            "OutIO2": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LVOutputConfig/OutIO2",
                106,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                ],
                None,
                None,
                "ALL",
            ),
            "OutIO4": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LVOutputConfig/OutIO4",
                107,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                ],
                None,
                None,
                "ALL",
            ),
            "LightInts": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LVOutputConfig/LightInts", 86, "ALL"
            ),
            "LightTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/LightTimeOut", 2, "ALL"
            ),
            "FiberTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/FiberTimeOut", 3, "ALL"
            ),
            "PumpTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/PumpTimeOut", 4, "ALL"
            ),
            "L120Timer": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/TimeOut/L120Timer",
                1,
                2,
                ["Shared", "Own"],
                None,
                2,
                "ALL",
            ),
            "QuietTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/QuietTimeOut", 63, "ALL"
            ),
            "FiltInterface": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/FiltInterface",
                27,
                None,
                ["StartDurFre", "NotUsed1", "NotUsed2", "PurgeOnly"],
                None,
                None,
                "ALL",
            ),
            "CleanP1": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/FilterConfig/CleanP1", 23, 0, "ALL"
            ),
            "CleanCP": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/FilterConfig/CleanCP", 24, 0, "ALL"
            ),
            "CPAlwaysON": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/FilterConfig/CPAlwaysON", 25, 0, "ALL"
            ),
            "FiltOTOption": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/FilterConfig/FiltOTOption", 26, 4, "ALL"
            ),
            "FiltCPOTOption": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/FilterConfig/FiltCPOTOption", 26, 5, "ALL"
            ),
            "FiltSuspendBy": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/FiltSuspendBy",
                26,
                6,
                ["AnyUD", "PumpUD"],
                None,
                2,
                "ALL",
            ),
            "FiltCPOTLimitOption": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/FiltCPOTLimitOption",
                26,
                7,
                ["AlwaysEnabled", "LimitedIfSPBelow95F"],
                None,
                2,
                "ALL",
            ),
            "PurgeSpeed": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/PurgeSpeed",
                26,
                0,
                ["Lo", "Hi"],
                None,
                2,
                "ALL",
            ),
            "FiltOTActTemp": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltOTActTemp",
                47,
                "ALL",
            ),
            "FiltOTDeadband": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltOTDeadband",
                48,
                "ALL",
            ),
            "FiltMinDur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/AdvanceFilterConfig/FiltMinDur", 49, "ALL"
            ),
            "FiltCPOTActTemp": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltCPOTActTemp",
                50,
                "ALL",
            ),
            "FiltCPOTDeadband": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltCPOTDeadband",
                51,
                "ALL",
            ),
            "FiltCPMinOnTime": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltCPMinOnTime",
                52,
                "ALL",
            ),
            "FiltCPOTMaxDur": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltCPOTMaxDur",
                53,
                "ALL",
            ),
            "FiltSuspendTime": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltSuspendTime",
                54,
                "ALL",
            ),
            "O3Pump": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/O3Config/O3Pump",
                55,
                None,
                ["CP", "P1"],
                None,
                None,
                "ALL",
            ),
            "O3FollowPump": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/O3Config/O3FollowPump", 56, 0, "ALL"
            ),
            "O3DuringFilt": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/O3Config/O3DuringFilt", 57, 0, "ALL"
            ),
            "O3SuspendBy": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/O3Config/O3SuspendBy",
                58,
                None,
                ["AnyUD", "PumpUD"],
                None,
                None,
                "ALL",
            ),
            "O3FilterDelay": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/O3Config/O3FilterDelay", 59, "ALL"
            ),
            "O3SuspendTime": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/O3Config/O3SuspendTime", 60, "ALL"
            ),
            "O3TogglePeriod": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/O3Config/O3TogglePeriod", 61, "ALL"
            ),
            "SaniLevel": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/Sanitation/SaniLevel", 45, "ALL"
            ),
            "SaniOnTime": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/Sanitation/SaniOnTime", 46, "ALL"
            ),
            "TempUnits": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/TempUnits",
                13,
                2,
                ["C", "F"],
                None,
                2,
                "ALL",
            ),
            "SetpointG": GeckoTempStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/SetpointG",
                15,
                "ALL",
            ),
            "HeaterPump": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/HeaterPump",
                12,
                None,
                ["CP", "P1"],
                None,
                None,
                "ALL",
            ),
            "HeatRestriction": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/HeatRestriction",
                85,
                3,
                ["No", "FullPowerOnly"],
                None,
                2,
                "ALL",
            ),
            "ExtProbeEnable": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/ExtProbeEnable",
                14,
                4,
                "ALL",
            ),
            "OverSetpointEnable": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/OverSetpointEnable",
                14,
                5,
                "ALL",
            ),
            "EconBelowSetpoint": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/EconBelowSetpoint",
                64,
                "ALL",
            ),
            "LineFrequency": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/LineFrequency",
                85,
                1,
                ["60HZ", "50HZ"],
                None,
                2,
                "ALL",
            ),
            "LineVoltage": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/LineVoltage",
                85,
                2,
                ["240Vac", "230VAC"],
                None,
                2,
                "ALL",
            ),
            "NbPhases": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/NbPhases", 87, None
            ),
            "BrNbSettings": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/BrNbSettings", 88, None
            ),
            "BrSetIndex": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/BrSetIndex", 89, None
            ),
            "BreakerSet1": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/BreakerSet1", 90, "ALL"
            ),
            "BreakerSet2": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/BreakerSet2", 91, "ALL"
            ),
            "BreakerSet3": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/BreakerSet3", 92, "ALL"
            ),
            "BreakerSet4": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/BreakerSet4", 93, "ALL"
            ),
            "BreakerSet5": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/BreakerSet5", 94, "ALL"
            ),
            "CE_2x20Allowed": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InputSupply/CE_2x20Allowed", 62, 6, "ALL"
            ),
            "UDActionOnQuiet": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/Quiet/UDActionOnQuiet",
                62,
                1,
                [
                    "IgnoreUD",
                    "ExitQuietAndRestoreAllUD",
                    "NA",
                    "ExitQuietAndCancelOtherUD",
                ],
                None,
                4,
                "ALL",
            ),
            "QuietActionOnUD": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/Quiet/QuietActionOnUD",
                62,
                0,
                ["Cancel", "Pause"],
                None,
                2,
                "ALL",
            ),
            "QuietActionOnHeater": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/Quiet/QuietActionOnHeater",
                13,
                3,
                ["NoAction", "ForceMaxPower", "KeepOff", ""],
                None,
                4,
                "ALL",
            ),
            "FanActiveOnAmbiantOT": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/Fan/FanActiveOnAmbiantOT", 13, 5, "ALL"
            ),
            "FanSetAOTTempAdc": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/Fan/FanSetAOTTempAdc", 17, "ALL"
            ),
            "FanClrAOTTempAdc": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/Fan/FanClrAOTTempAdc", 18, "ALL"
            ),
            "FanActiveOnPumpRun": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/Fan/FanActiveOnPumpRun", 13, 6, "ALL"
            ),
            "FanActiveDelayAfterPumpRun": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/Fan/FanActiveDelayAfterPumpRun", 19, "ALL"
            ),
            "FanForceOFFOption": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/Fan/FanForceOFFOption", 13, 7, "ALL"
            ),
            "FanSetForceOffTempAdc": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/Fan/FanSetForceOffTempAdc", 20, "ALL"
            ),
            "FanClrForceOffTempAdc": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/Fan/FanClrForceOffTempAdc", 21, "ALL"
            ),
            "FanForceOffInhibitDelay": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/Fan/FanForceOffInhibitDelay", 22, "ALL"
            ),
            "FanPump": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/Fan/FanPump",
                14,
                1,
                ["None", "P1", "CP"],
                None,
                4,
                "ALL",
            ),
            "FanAsCoolingDevice": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/Fan/FanAsCoolingDevice", 14, 3, "ALL"
            ),
            "Pump1AsVSP": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/VSP/Pump1AsVSP", 5, 0, "ALL"
            ),
            "Pump3AsVSP": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/VSP/Pump3AsVSP", 5, 2, "ALL"
            ),
            "VSPCheckfloSpeed": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/VSP/VSPCheckfloSpeed", 6, "ALL"
            ),
            "VSPFilterSpeed": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/VSP/VSPFilterSpeed", 7, "ALL"
            ),
            "VSPSpeedLevel0": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/VSP/VSPSpeedLevel0", 8, "ALL"
            ),
            "VSPSpeedLevel1": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/VSP/VSPSpeedLevel1", 9, "ALL"
            ),
            "VSPSpeedLevel2": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/VSP/VSPSpeedLevel2", 10, "ALL"
            ),
            "VSPSpeedLevel3": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/VSP/VSPSpeedLevel3", 11, "ALL"
            ),
            "BlowerKeyOption": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/BlowerKeyOption",
                1,
                0,
                ["LastPumpKey", "NextFreePumpKey"],
                None,
                2,
                "ALL",
            ),
            "QuickOnOffKeyEnable": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/QuickOnOffKeyEnable",
                1,
                1,
                "ALL",
            ),
            "K600ForMay": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/K600ForMay",
                62,
                7,
                "ALL",
            ),
            "FiltFreq": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/FiltFreq",
                28,
                "ALL",
            ),
            "FiltStart": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/FiltStart",
                29,
                "ALL",
            ),
            "FiltDur": GeckoTimeStructAccessor(
                self.struct, "ConfigStructure/SpaPackInternalOptions/FiltDur", 31, "ALL"
            ),
            "OnzenStart": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/OnzenStart",
                37,
                "ALL",
            ),
            "OnzenDur": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/OnzenDur",
                39,
                "ALL",
            ),
            "OnzenFreq": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/OnzenFreq",
                41,
                "ALL",
            ),
            "EconStart": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconStart",
                65,
                "ALL",
            ),
            "EconStop": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconStop",
                67,
                "ALL",
            ),
            "EconControlableManually": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconControlableManually",
                62,
                3,
                "ALL",
            ),
            "EconProgStatus": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconProgStatus",
                62,
                4,
                ["Disabled", "NotActive", "NA", "Active"],
                None,
                4,
                "ALL",
            ),
            "K600Msg1": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/K600Reminders/K600Msg1",
                69,
                None,
                [
                    "CleanFilter",
                    "PH",
                    "CheckGFCI",
                    "ChangeWater",
                    "CleanCover",
                    "ReplaceFilter",
                ],
                None,
                None,
                "ALL",
            ),
            "K600Msg2": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/K600Reminders/K600Msg2",
                70,
                None,
                [
                    "CleanFilter",
                    "PH",
                    "CheckGFCI",
                    "ChangeWater",
                    "CleanCover",
                    "ReplaceFilter",
                ],
                None,
                None,
                "ALL",
            ),
            "K600Msg3": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/K600Reminders/K600Msg3",
                71,
                None,
                [
                    "CleanFilter",
                    "PH",
                    "CheckGFCI",
                    "ChangeWater",
                    "CleanCover",
                    "ReplaceFilter",
                ],
                None,
                None,
                "ALL",
            ),
            "K600Msg4": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/K600Reminders/K600Msg4",
                72,
                None,
                [
                    "CleanFilter",
                    "PH",
                    "CheckGFCI",
                    "ChangeWater",
                    "CleanCover",
                    "ReplaceFilter",
                ],
                None,
                None,
                "ALL",
            ),
            "K600Msg5": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/K600Reminders/K600Msg5",
                73,
                None,
                [
                    "CleanFilter",
                    "PH",
                    "CheckGFCI",
                    "ChangeWater",
                    "CleanCover",
                    "ReplaceFilter",
                ],
                None,
                None,
                "ALL",
            ),
            "K600Msg6": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/K600Reminders/K600Msg6",
                74,
                None,
                [
                    "CleanFilter",
                    "PH",
                    "CheckGFCI",
                    "ChangeWater",
                    "CleanCover",
                    "ReplaceFilter",
                ],
                None,
                None,
                "ALL",
            ),
            "K600Msg7": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/K600Reminders/K600Msg7",
                75,
                None,
                [
                    "CleanFilter",
                    "PH",
                    "CheckGFCI",
                    "ChangeWater",
                    "CleanCover",
                    "ReplaceFilter",
                ],
                None,
                None,
                "ALL",
            ),
            "K600Msg8": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/K600Reminders/K600Msg8",
                76,
                None,
                [
                    "CleanFilter",
                    "PH",
                    "CheckGFCI",
                    "ChangeWater",
                    "CleanCover",
                    "ReplaceFilter",
                ],
                None,
                None,
                "ALL",
            ),
            "K600Msg1NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg1NbWeek", 77, "ALL"
            ),
            "K600Msg2NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg2NbWeek", 78, "ALL"
            ),
            "K600Msg3NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg3NbWeek", 79, "ALL"
            ),
            "K600Msg4NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg4NbWeek", 80, "ALL"
            ),
            "K600Msg5NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg5NbWeek", 81, "ALL"
            ),
            "K600Msg6NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg6NbWeek", 82, "ALL"
            ),
            "K600Msg7NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg7NbWeek", 83, "ALL"
            ),
            "K600Msg8NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg8NbWeek", 84, "ALL"
            ),
            "TimeFormat": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/TimeFormat",
                85,
                0,
                ["AmPm", "24h"],
                None,
                2,
                "ALL",
            ),
            "AmbiantOTEnable": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/AmbiantOTEnable", 14, 0, "ALL"
            ),
        }
